from django.forms import Form
from django.test import TestCase
from django.urls import reverse
from django.contrib.auth import get_user_model

User = get_user_model()

class LoginCheckTestCase(TestCase):
    """Basic tests of the login check endpoint"""
    
    def setUp(self) -> None:
        """Get the url and create a user"""
        self.url = reverse("django_accounts_api:login_check")
        self.user = User.objects.create_user("test")
        return super().setUp()

    def testGetLoginNotAuthenticated(self):
        """Just hitting this endpoint with no login should return 401"""
        response = self.client.get(self.url)
        assert response.status_code == 401

    def testGetLoginAuthenticated(self):
        """hitting this endpoint with login should return 200"""
        self.client.force_login(self.user)
        response = self.client.get(self.url)
        assert response.status_code == 200
        assert "name" in response.json()


class LogoutTestCase(TestCase):
    """Basic tests of the logout endpoint"""
    
    def setUp(self) -> None:
        """Get the url and create a user"""
        self.url = reverse("django_accounts_api:logout")
        self.user = User.objects.create_user("test")
        self.client.force_login(self.user)
        return super().setUp()

    def testGetLogout(self):
        """Just getting this endpoint with no login should return 200"""
        response = self.client.get(self.url)
        assert response.status_code == 200

    def testPostLogout(self):
        """posting to this endpoint with login should return 200"""
        test = self.client.force_login(self.user)
        response = self.client.post(self.url)
        assert response.status_code == 200


class LoginViewTestCase(TestCase):
    """Basic tests of the partial HTML login view"""
    
    def setUp(self) -> None:
        """Get the url and create a user"""
        self.url = reverse("django_accounts_api:login")
        self.user = User.objects.create_user("test", password="test")
        return super().setUp()

    def testGetLogin(self):
        """An unauthed get should return a form"""
        response = self.client.get(self.url)
        assert response.status_code == 200
        assert "form" in response.context
    
    def testPostLogin(self):
        """A post without username and password should display errors"""
        response = self.client.post(
            self.url, {})
        assert response.status_code == 200
        assert "form" in response.context
        form: Form = response.context["form"]
        assert 'username' in form.errors
        assert 'password' in form.errors
    
    def testPostLoginNoPassword(self):
        """A post without password should display an error"""
        response = self.client.post(
            self.url, dict(
                username='test'
            ))
        assert response.status_code == 200
        assert "form" in response.context
        form: Form = response.context["form"]
        assert 'password' in form.errors
        assert len(form.errors) == 1

    def testPostLoginIncorrect(self):
        """A post with incorrect credentials should display an error"""
        response = self.client.post(
            self.url, dict(
                username='test',
                password="bad"
            ))
        assert response.status_code == 200
        assert "form" in response.context
        form: Form = response.context["form"]
        assert '__all__' in form.errors
        assert len(form.errors) == 1

    def testPostLoginCorrect(self):
        """A post with incorrect credentials should return a 201"""
        response = self.client.post(
            self.url, dict(
                username='test',
                password="test"
            ))
        assert response.status_code == 201
        assert "name" in response.json()


class LoginFlowTestCase(TestCase):
    """A flow test of the login endpoints"""
    
    def setUp(self) -> None:
        """Get the url and create a user"""
        self.logout_url = reverse("django_accounts_api:logout")
        self.check_url = reverse("django_accounts_api:login_check")
        self.url = reverse("django_accounts_api:login")
        self.user = User.objects.create_user("test", password="test")
        return super().setUp()
    
    def testFlow(self):
        """ user should be unauthed, login, then be authed, then hit log out, then be unauthed"""
        response = self.client.get(self.check_url)
        assert response.status_code == 401
        response = self.client.post(
            self.url, dict(
                username='test',
                password="test"
            ))
        assert response.status_code == 201
        response = self.client.get(self.check_url)
        assert response.status_code == 200
        response = self.client.post(self.logout_url)
        assert response.status_code == 200
        response = self.client.get(self.check_url)
        assert response.status_code == 401


from django.contrib.auth.models import User as DjangoUser
from django.contrib.auth.views import LoginView, LogoutView, PasswordChangeView
from django.http import HttpRequest, HttpResponse, JsonResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters


def manifest(request: HttpRequest) -> HttpResponse:
    return JsonResponse(dict(
        check=reverse("django_accounts_api:login_check"),
        login=reverse("django_accounts_api:login"),
        logout=reverse("django_accounts_api:logout"),
        password_change=reverse("django_accounts_api:password_change"),
    ))

def _user_details(user: DjangoUser) -> dict:
    """The details of the user to return on success"""
    return dict(
        id=user.pk,
        name=user.get_full_name(),
    )

def login_check(request) -> HttpResponse:
    """200 and details if logged in, 401 if not"""
    user: DjangoUser = request.user
    if(user.is_authenticated):
        return JsonResponse(_user_details(user))
    else:
        return HttpResponse(status=401)


class LoginPartialHTML(LoginView):
    ''' Override the Django login view to NOT redirect on successful login 
    GET - renders a partial login form
    POST - submits and validates
        if invalid returns form rendered with errors
        if valid logs the user in
    '''

    template_name = "django_accounts_api/login.html"
    def form_valid(self, form):
        """Override redirect behaviour"""
        _repressed_redirect = super().form_valid(form)
        return JsonResponse(
            _user_details(self.request.user),
            status=201
        )


class ApiLogout(LogoutView):
    ''' Override the Django logout view to NOT redirect on successful login 
    GET - actually calls POST, but will error in Django 5
    POST - logs out, returns 200
    '''

    def post(self, request, *args, **kwargs):
        _repressed_redirect_or_render = super().post(request, *args, **kwargs)
        return HttpResponse(
            status=200
        )


@method_decorator(sensitive_post_parameters(), name='dispatch')
@method_decorator(csrf_protect, name='dispatch')
class ChangePasswordPartialHTML(PasswordChangeView):
    ''' Override the Django change password view to support API use
    GET - renders a partial change password form - can be accessed without auth
    '''
    template_name = "django_accounts_api/password_change.html"
    
    def dispatch(self, *args, **kwargs):
        return super(PasswordChangeView, self).dispatch(*args, **kwargs)
    
    def post(self, *args, **kwargs):
        if self.request.user.is_authenticated:
            return super().post(*args, **kwargs)
        else:
            return HttpResponse(status=401)
    
    def form_valid(self, form):
        """Override redirect behaviour"""
        _repressed_redirect = super().form_valid(form)
        return JsonResponse(
            _user_details(self.request.user),
            status=201
        )


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_accounts_api', 'django_accounts_api.migrations']

package_data = \
{'': ['*'], 'django_accounts_api': ['templates/django_accounts_api/*']}

install_requires = \
['django>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'django-accounts-api',
    'version': '0.1.3',
    'description': '',
    'long_description': "# Django Api Tools\n\nA collection of apps that support basic django functionality over api\n\nCaveat enptor, very early days, still being tested in it's first project\n\n# Use\n\n- `pip install ...`\n- add `'django_accounts_api',` to INSTALLED_APPS\n- add `path('/accounts_api/', include('django_accounts_api.urls'))` to your urls\n\n## Features\n\n### API endpoints\n\nScenario to support is a compiled javascript capable frontend needing to provide authentication features over api\nFrontend should be able to:\n\nOptionally get api endpoints from api.\n\n| Task | How |\n|-----|-----|\n| find api endpoints | GET `/manifest` |\n| Detect auth | GET `/login-check` |\n| show user name | json parse GET `/login-check` response |\n| render a login form | GET `/login` body returns partial HTML |\n| submit a login form | wrap partial form above in a form, `new FormData(...)` it and POST to `/login |\n| render form errors | render partial HTML returned from POST `/login` |\n| login | POST `/login` returns 201 |\n| logout | POST `/logout` |\n\n\n## Development\n1. Install Poetry https://python-poetry.org/docs/#installation\n\n2. Use a virtual environment https://python-poetry.org/docs/basic-usage/#using-your-virtual-environment\n\n3. `poetry install --with test,dev --no-root`\n\n4. `pytest`\n",
    'author': 'PeteCoward',
    'author_email': 'peter@catalpa.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
